<?php
// $Id: transport.remote_entity.inc,v 1.1.2.15 2011/02/24 11:50:36 joachim Exp $
/**
 * @file transport.remote_entity.inc
 * Contains the transport remote entity class, subclasses, and interfaces.
 */

/**
 * The base class to handle a remote entity.
 *
 * The lifecycle of a remote entity runs through several phases.
 * Each phase is run with a standard method, of the form run_phase_PHASENAME.
 *
 * Phase methods are defined in the following places:
 *  - the base class
 *  - the class for a particular entity type
 *  - the interface extenders for a particular combination of entity type AND source type.
 * This does cause a bit of jumping around the code, but with clear commenting it's not too bad ;)
 *
 * Each phase may choose to hand over control to the next phase or halt the 
 * entire processing of the entity (and thus return to the controller) 
 * -- currently only the dependencies phase does this.
 *
 * We faces.module's extender faces pattern to allow phase methods to be
 * defined in any of:
 *  - this base class
 *  - a child class for the entity type. 
 *    Class names are of the form TransportRemoteEntity_TYPE.
 *  - An interface loaded by the chid class. This particular pattern allows
 *    us to define a phase method for any combination of entity type and
 *    source type. This is what allows different ways to fetch remote data
 *    depending on entity type and source type (eg, getting node data from
 *    a Drupal 5 site, getting user data from a Drupal 6 site, etc).
 *
 * The phases are as follows:
 *
 * - object construction (not really a phase)
 *   - The object is created. Its class reflects the type of entity.
 *   - The object is extended further according to source and type.
 *     TODO: create a system allowing applications to request a subset
 *     of phases, or to add extra phases.
 *   - The local entity is identified, if one exists.
 *   - The stash is checked: if we already fetched this object once in this set,
 *     we are free to skip the fetch phase.
 *     TODO: circular dependencies -- need to think about how to handle these.
 *
 * 'init'
 *    Currently not much, but this is where entity can declare that it should be
 *    skipped. Eg, users with uid 0 and 1 are always skipped. 
 *    TODO: at this point we should check to see if the entity has been previously 
 *    retrieved within a set timeframe, eg if the node was retrieved less than a day ago
 *    consider it to still be fresh and skip it.
 *    Freshness should be per-type; eg you might want to say that taxonomy terms
 *    never need to be retrieved again and should always be skipped.
 *    TODO: consider moving things from the constructor to this phase.
 * 
 * 'fetch'
 *    - The data for the entity is fetched from the source.
 *    - The entity is compared with the local version of the entity, if it exists.
 *    - TODO: If this is more recent than the required freshness of the controller operation,
 *      the entity is skipped; control returns to the Controller with a status to move on to the next entity.
 * 
 * 'filter'
 *   The entity is run through:
 *     1. Custom filters provided by drupal_alter hooks.
 *       This is where you can opt to remove properties that are not relevant to the local site, or
 *       massage data. 
 *     2. The filter provided by its type, according to source type.
 *       Eg, taxo terms from Drupal 5 automatically get tidied up for the 
 *       current local version of Drupal.
 * 
 * 'dependencies'
 *   The entity is run through a hook that allows all modules to add to a list
 *   of entity dependencies. This is a flat list, eg term 35, node 4, user 3.
 *   At this point, if there are dependencies, execution of the entity handler
 *   stops and control goes back to the controller object. The controller
 *   processes the dependent entities first, and then resumes this entity's
 *   execution at its next phase.
 *   
 * 'translate'   
 *   References in the entity to remote entities are translated to local IDs.
 *   Eg, a node's author is set to a local user (who has been retrieved as a
 *   dependency).
 *
 * 'preexisting' -- TODO
 *   If the entity has not previously been retrieved, check whether an entity
 *   exists locally which matches it anyway. Eg, check for a local user with
 *   the same email address.
 * 
 * 'save'
 *   The entity is saved.
 *
 * 'record'
 *   The correspondence between remote ID and local ID for this particular
 *   source is stored in the database.
 */
class TransportRemoteEntity extends FacesExtendable {
  // These pretty much all have to be public because objects of this class
  // need to be serialized to the database. Bang goes our privacy :(
  // Hence there's not that much reason to declare them upfront, apart from
  // documentation and trying to prevent key-itis -- listing used object keys
  // may help to spot redundancies and prevent clutter.

  // TODO: phases should get set by the constructor so 
  // more that defined here can be provided.
  // ALSO: need to distinguish phases that take an interface and those that don't!
  // Some phases are commented out and TODO, as I don't need them for this current project.
  public $phases = array(
    'init',
    'fetch',
    'filter',
    'dependencies',
    'translate',
    'mappings',
    // 'preexisting', // TODO
    'save',
    'record'
  );
  // Phases which need to load interfaces. 
  // Graaargh I don't like the partial duplication with the above $phases, but
  // the alternative is a more complex array. One to ponder. TODO
  // In fact, consider naming faces that are not phase names,
  // as we only need one face in the core module at least.
  // Hence this one face could be called 'core', or 'source' since it
  // diverges on source type.
  public $faced_phases = array(
    'fetch', 
    //'filter',
  );
  
  // Lifecycle properties.
  public $current_phase;   // The phase currently being run or just completed.
  public $phase_checklist; // This array of phases is emptied as each one is run.
  public $status;          // Holds a string describing current status.
    // Possible values include:
    //  - run: handler is running through phases
    //  - error
    //  - complete
    //  - dependencies: handler wants to hand back to the controller for dependencies
    //  - skip: handler wants to hand back to the controller and skip this entity entirely.
  public $advance_phase; // TRUE as long as the handler should continue to work
    // through phases.
    // Phases may set this to FALSE to halt operation.
    // They should ALSO set a status.
    // TODO: phases probably only need to set $this->status to something != 'run'.
    // TODO: refactor this into a method setStatus that's easier (and more rigid) for
    // methods to call, eg setStatus('error'), and the method also takes care of
    // setting advance_phase to FALSE.
  
  // Data about the entity this object handles.
  public $data; // The original data from the controller of what to retrieve.
  public $type;
  public $id_remote;
  public $result; // This holds exactly what we fetched from remote and is not changed.
  public $entity; // This holds the data we fetch from remote
    // and is gradually manipulated over the lifecycle.
  public $module_data;
    // Modules may stash anything they like in this array, keyed by module name.
  
  // Data about the local entity.
  public $local_exists = FALSE; // Flag whether we have imported in a previous set. Does not change.
    // If this is TRUE, the $this->id_local also is set.
  public $prior_exists = FALSE; // Flag whether this object exists but has not been retrieved by us.
    // Example: a user with the same email address.
  public $local_id; // The local id of the entity. Set as soon as the entity is saved.

  // Data from the controller: see the controller class for more on these.
  public $controller; // The controller object
  public $base_time;  // The controller's base time of execution. This identifies a set.
  public $source_name;
  public $source;     // The source object. So far, a Clients class.
  
  /**
   * Constructor method.
   *
   * Set properties for the object and loads interfaces.
   */
  function __construct($controller, $entity_data) {
    // Set our controller and copy some data from it.
    $this->controller   = $controller;
    $this->source_name  = $controller->source_name;
    $this->source       = $controller->source;
    $this->base_time    = $controller->get_base_time();
    
    //dsm("basetime: $this->base_time");
    
    // Set our basic parameters from what we were passed in.
    $this->data       = $entity_data;
    $this->type       = $entity_data['type'];
    $this->id_remote  = $entity_data['id']; // BAD NAME. CHANGE.
    $this->id_local;
    
    // Extend the class with interfaces.
    // For each *requested* phase (API TODO!), extend 

    //dsm($this->controller->remote_entity_classes);
    $source_class = get_class($this->source);
  
    // Load the interface for each phase.
    foreach ($this->faced_phases as $phase) {
      //dsm($phase);
      //dsm($this->controller->remote_entity_classes[$this->type]['phases']);
      $interface = 'TransportRemoteEntityInterface_' . $phase;
      $extender_class = $this->controller->remote_entity_classes[$this->type]['phases'][$phase][$source_class];
      //dsm($interface_class);
      $this->extendByClass(array($interface), $extender_class);
    }
    
    // Make a copy of the phase list that we can shift off to keep track of what's next.
    $this->phase_checklist = $this->phases;
    
    // Mark that we're new so drupal_write_record() can be told later
    // when stashing the object between phases.
    $this->new = TRUE;
    
    // Find out if this entity has previously been retrieved.
    $this->get_local_entity();
  }
  
  /**
   * Due to a WTF in Faces we have to implement this ourselves :/
   */
  public function __sleep() {
    $array = get_object_vars($this);
    return array_keys($array);
  }  
  
  /**
   * Accessor functions.
   */
  function get_id() {
    return $this->id_remote;
  }
  
  /**
   * Get the local id of the remote entity we represent, in the event that
   * it's previously been retrieved.
   */
  function get_local_entity() {
    //dpm("get local $this->source_name, $this->source_id, $this->type");
    $result = db_query("SELECT id FROM {transport_entity_remote} WHERE source = '%s' AND id_remote = %d AND type = '%s'",
      $this->source_name, $this->id_remote, $this->type);
      
    $local_entity_id = (int) db_result($result);
    if ($local_entity_id) {
      $this->id_local     = $local_entity_id;
      $this->local_exists = TRUE;
      //dpm("local exists: $this->type $local_entity_id");
    }
  } 
  
  /**
   * Run the object through as many phases as possible.
   *
   * The controller object calls this. Operation is handed back to it either
   * when the entity is completely retrieved, or it encounters a problem or
   * a dependency to resolve.
   */
  function run() {
    // Loop through phases.
    // Each phase has an entry point function run_phase_FOO.
    // This should set the property $this->advance_phase to:
    //  TRUE if the phase is complete
    //  FALSE otherwise,
    // in which case we need to look into the status of the object for more details.
    $this->advance_phase = TRUE;
    $this->status = 'run';
    //dsm($this->phase_checklist);
    while ($this->advance_phase && count($this->phase_checklist)) {
      
      $this->current_phase = array_shift($this->phase_checklist);
      $this->phase = $this->current_phase; // ugly :(

      // Run the current phase.
      $phase_function = 'run_phase_' . $this->current_phase;
      //dsm("phase $this->current_phase -- function: $phase_function");
      $this->$phase_function();
      
      //$this->advance_phase = FALSE;
      
      // For now, just stash at the end of every phase.
      // TODO: not all phases require stashing; really only those that may halt operation.
      // TODO: bug! If a given entity is a dependency multiple times, then we may
      // get here with the same time, type, id, AND phase and cause a duplicate error 
      // on the database!!!!
      $this->object = $this;
      if (isset($this->new)) {
        drupal_write_record('transport_processing_entities', $this);
        unset($this->new);
      }
      else {
        drupal_write_record('transport_processing_entities', $this, array('base_time', 'type', 'id_remote'));
      }
      unset($this->object);
    } // while loop.
    
    //dsm("left run at phase $this->current_phase status $this->status");
    
    if ($this->status == 'run') {
      // If we've stopped executing phases, but everything is still set to run
      // then it's because we've completed everything.
      if ($this->advance_phase) {
        $this->complete = TRUE;
        $this->status = 'complete';
      }
      else {
        // We should never get here: phases should set a status at the same time
        // as setting advance_phase to FALSE.
        $this->stalled = TRUE;
        $this->status = 'stalled';
      }
    }
  }
  
  /**
   * Phase handler: init - common.
   *
   * Just an empty stub to be overridden for now.
   * TODO: see if anything can be usefully moved here from the contructor method.
   */
  function run_phase_init() {
  }
  
  /**
   * Phase handler: dependencies - common.
   *
   * This should be common to all entities.   
   *
   * Now the object has been filtered we don't need to worry about versions of
   * Drupal: assume it's in a format that could be saved directly.
   *
   * However we may need other objects to exist locally (and have IDs) before
   * we save the current one.
   *
   * This phase allows modules to add dependencies.
   *
   * TODO: add a system to allow the application layer to skip some or
   * all dependencies. Using the filter phase to just delete keys won't
   * cut it, as those may need to be saved anyway.
   * Use case is migration where the order of operations can be fixed in 
   * a logical order.
   * Idea: set a $dep_skip_keys array in the initial $options,
   * and in this phase, stash those keys before invoking the hook,
   * then put them back straight after.
   */  
  function run_phase_dependencies() {
    // Invoke hook_transport_process_dependencies_TYPE() in all modules to
    // let them process the entity and add to its dependency list
    //$dependencies = module_invoke_all('transport_process_dependencies_' . $this->type, $this->entity);
    
    $dependencies_hook = 'transport_process_dependencies_' . $this->type;
    foreach (module_implements($dependencies_hook) as $module) {
      $dependencies_by_module[$module] = module_invoke($module, $dependencies_hook, $this);
      // Allow hooks to be lazy and return nothing.
      if (is_null($dependencies_by_module[$module])) {
        $dependencies_by_module[$module] = array();
      }
    }
    //dsm($dependencies_by_module);
    
    // We let implementing modules just return a flat, numeric array where
    // each element has the keys:
    //  - type: the entity type, eg node, user, term
    //  - id: the entity id.
    //  - ref: a variable reference to the property on the entity that stores 
    //      the dependency ID. Eg for the node author, this is set to
    //      &$node->uid.
    //      This bit is all quite scary... see the translate phase for more detail.
    // This is to provide a nice, uncomplicated (!) API that doesn't involve
    // making your data jump through hoops.
    // Now we rearrange it extensively...
    
    $dependencies_flat = array();
    foreach ($dependencies_by_module as $module => $module_data) {
      foreach ($module_data as $dep_data) {
        // Create a key in the standard form we use of TYPE_ID.
        // TODO: consider changing the separator to - since that will never
        // occur in machine names!
        $key = $dep_data['type'] . '_' . $dep_data['id'];
        // Index the new array by the key, and add in some extra values on the
        // dependency data item.
        $dependencies_flat[$key] = $dep_data;
        $dependencies_flat[$key]['module']  = $module;
        $dependencies_flat[$key]['class']   = 'TransportRemoteEntity_' . $dep_data['type'];
      }
    }
    //dsm($dependencies_flat);
    
    // Now we have all dependencies.
    $this->dependencies = $dependencies_flat;
    
    // If there are dependencies, stop the process on this entity;
    // return to the controller object.
    if (count($this->dependencies)) {
      $this->advance_phase = FALSE;
      $this->status = 'dependencies';
    }
  }
  
  /**
   * Phase handler: translate - common.
   *
   * Translate dependencies recorded by the dependencies phase.
   */  
  function run_phase_translate() {
    //dsm($this->dependencies, 'deps translate');
    
    
    // This works by magic.
    // The dependency hooks stored in the deps array a variable reference to the
    // actual variable on the entity that has the remote ID. 
    // We can then just grab this from the deps array and change it, and the entity
    // magically changes too!
    // Eg, we have $deps_key = &$user->uid, and here we change $deps_key.
    // Credit to http://stackoverflow.com/questions/4957232/is-there-way-for-programmatically-storing-an-address-of-properties-for-an-objec
    // Big question though: will it still work if the object is stashed in the database and
    // retrieved on a later run? 
    // It seems to!!!!
    foreach ($this->dependencies as $key => $dep) {
      $id_local  = transport_get_local_id($this->source_name, $dep['type'], $dep['id']);
      //dsm($id_local);
      $this->dependencies[$key]['ref'] = $id_local;
    }
    //$id_local  = transport_get_local_id($entity_handler->source_name, 'user', $remote_entity->uid);
    //$remote_entity->uid = $uid_local;
  }
  
  /**
   * Phase handler: mappings - common.
   *
   * Convert global remote ids to local ids and other miscellaneous mappings;
   * currently just input formats.
   */  
  function run_phase_mappings() {   
    // Get mappings
    // TODO: cache this!
    // Invoke hook_transport_mappings_list() in all modules to build up an
    // array of id mappings.
    // This an array of mapping types, and each of these is an array of 
    // remote type to local type.
    // Eg:
    //  'input_format' => array(
    //    REMOTE_ID => LOCAL_ID
    //  ),
    //  'another_type' => etc
    // TODO: ugly variable names!
    $mappings = module_invoke_all('transport_mappings_list', $this->entity);
    //dsm($mappings, 'mappings');
    
    // This gets us a list of properties that need mapping.
    // TODO: consider folding the gathering of this in with the dependencies phase
    // and then 'dependencies' is just a not terribly special group of mapping
    // and so it can all be done in the translate phase or something...
    $mappings_hook = 'transport_process_mappings_' . $this->type;
    $properties = module_invoke_all($mappings_hook, $this);
    //dsm($properties);
    
    foreach ($properties as $property) {
      $group = $property['group'];
      if (isset($mappings[$group])) {
        $id_remote = $property['id'];
        if (isset($mappings[$group][$id_remote])) {
          $property['ref'] = $mappings[$group][$id_remote];
        }
      }
    }
    
    
    //dsm($this->entity, 'done mappings');
  }

  /**
   * Phase handler: record - common.
   *
   * This should be common to all entities.
   *
   * TODO: only record if something was saved!!!!!
   */
  function run_phase_record() {  
    // Only record if no local id.
    $record = array(
      'source'    => $this->source_name,
      'id_remote' => $this->id_remote,
      'type'      => $this->type,
      'base_time' => $this->base_time,
      'id'        => $this->id_local,
    );
    
    //dsm("recording $this->type $this->id_remote locally as  $this->id_local");
    
    // Test on local_exists, as the id_local will always be set by this point.
    // local_exists represents the state on initialization and does not
    // subsequently change.
    if ($this->local_exists) {
      drupal_write_record('transport_entity_remote', $record, array('source', 'type', 'id_remote'));
    }
    else {
      drupal_write_record('transport_entity_remote', $record);
    }
  }
}

// ======================================== Phase interfaces

/**
 * Some phases the remote entity can go through are handled by an interface.
 
 A thought -- do we really need interfaces for EVERY phase?
 how about just one interface for stuff that is source-specific??
 Let's try sticking it all in the 'fetch' interface until we hit trouble... 
 then if it works, just rename ;)
*/

/**
 * Interface for fetch phase.
 */
interface TransportRemoteEntityInterface_fetch {
  function run_phase_fetch();
  function run_phase_filter();
}

/**
 * Common class for fetch phase. Provides helper functions.
 */
abstract class TransportRemoteEntityExtender_fetch 
  extends FacesExtender 
  implements TransportRemoteEntityInterface_fetch {
}

/**
 * Common class for fetch phase on Client sources. Provides helper functions.
 */
abstract class TransportRemoteEntityExtender_fetch_client 
  extends TransportRemoteEntityExtender_fetch
  implements TransportRemoteEntityInterface_fetch {
  
  /**
   * Check the result of a client XMLRPC method call and handle errors.
   *
   * @return
   *  TRUE if the result is successful; FALSE otherwise.
   */
  function fetch_verify_client_result($result) {
    if (isset($result->is_error)) {
      // Halt execution of the handler and set the errors for retrieval by the controller.
      $this->object->advance_phase = FALSE;
      $this->object->status = 'error';
      $this->object->errors['fetch'] = $result;      
      return FALSE;
    }
    return TRUE;
  }
}

// ======================================== Stand-in hook implementations: common
/**
 * Implementation of hook_transport_mappings_list() for filter module.
 *
 * @return
 *  An array of global mappings, in the form:
 *  $mappings[GROUP] = array(
 *    REMOTE_ID => LOCAL_ID,
 *  );
 */
/* 
// ARGH: this is really lousy, as it makes absolutely no allowance for 
// different input formats on different remote sites.
// Need to ponder how to handle this for different applications that may
// use transport with several remote sites.
// In the meantime, this is an example of how to get input formats to
// be mapped in your own use of Transport -- implement this hook and 
// return the 'input_format' group as below.
// @see http://drupal.org/node/1064572 for more background.
function filter_transport_mappings_list() {
  return array(
    'input_format' => array(
      // remote_format_id => local_format_id,
      3 => 1,
    ),
  );  
}
*/

// ======================================== Extender classes: Node

/**
 * Entity class for nodes.
 *
 * The entity class should contain methods for phases that do not have interfaces;
 * ie those that are identical no matter the source.
 */
class TransportRemoteEntity_node extends TransportRemoteEntity {
  /**
   * Phase handler: node - init.
   */
  function run_phase_init() {
    // It is extremely important to call the parent method here.
    parent::run_phase_init();
  }

  /**
   * Phase handler: node - save.
   */  
  function run_phase_save() {
    //dsm($this->entity, 'about to save node');
    
    // If the node exists locally, set the nid and vid so node_save() performs an update.
    if ($this->local_exists) {
      //dsm("node exist LOCALLY - $this->id_local");
      $node_local = node_load($this->id_local);
      
      //dsm($node_local, '$node_local');
      
      $this->entity->nid = $node_local->nid;
      $this->entity->vid = $node_local->vid;
    }
    
    node_save($this->entity);
    $this->id_local = $this->entity->nid;
  }
}

/**
 * Extender Class: fetch phase.
 *
 * Type:   Node.
 * Source: Clients Drupal 5
 */
class TransportRemoteEntityExtender_fetch_node_d5 
  extends TransportRemoteEntityExtender_fetch_client 
  implements TransportRemoteEntityInterface_fetch {

  /**
   * Phase handler: node - fetch.
   *
   * The fetch phase retrieves data from the source and stores it in the entity object.
   */
  function run_phase_fetch() {
    // Cast to int for fussiness in XMLRPC.
    $nid = (int) $this->object->get_id();
    //dsm($nid);
    //dsm($this->object);
    //$result = $drupal_5->callMethod('user.load', 1);
    // Long chain of objects: because we're in a face, we want the object;
    // and its source is a Clients object on which we call the method to call a method ;)
    $result = $this->object->source->callMethod('node.load', $nid);
    if ($this->fetch_verify_client_result($result)) {
      $this->object->result = $result;
      $this->object->entity = (object) $result;
    }
    // If the result is an error, fetch_verify_client_result() takes care of
    // setting the remote entity handler status to 'error'.

    // Check for sanity here?
    //dsm($result, 'fetch result');
    //return "node from drupal! $nid";
  }
  
  /**
   * Phase handler: node - filter.
   *
   * The filter phase performs any clean-up required according to the object's version.
   */
  function run_phase_filter() {    
    // It's up to the individual implementation to translate any field names.
    // For instance, if your D5 site has an image in field_foo
    // and you want this to go into foo_bar, you need to move things
    // on the object using the alter hook.
    // Invoke hook_transport_TYPE_prefilter_alter().
    $type = $this->object->type;
    drupal_alter('transport_' . $type . '_prefilter', $this->object);

    // Actually very little needs to be done here usually.
    // Unset the nid and fid.
    unset($this->object->entity->nid);
    unset($this->object->entity->vid);
    
    
    // DEVELOPMENT CODE!!!!! REMOVE THIS LATER TODO TODO TODO
    unset($this->object->entity->taxonomy);
  }
}

/**
 * Extender Class: fetch phase.
 *
 * Type:   Node.
 * Source: Clients Drupal 6 Services 2
 */
class TransportRemoteEntityExtender_fetch_node_d6_2 
  extends TransportRemoteEntityExtender_fetch_client 
  implements TransportRemoteEntityInterface_fetch {

  /**
   * Phase handler: node - fetch.
   *
   * The fetch phase retrieves data from the source and stores it in the entity object.
   */
  function run_phase_fetch() {
    // Cast to int for fussiness in XMLRPC.
    $nid = (int) $this->object->get_id();
    //dsm($nid);
    //dsm($this->object);
    // Long chain of objects: because we're in a face, we want the object;
    // and its source is a Clients object on which we call the method to call a method ;)
    $result = $this->object->source->callMethod('node.get', $nid);
    if ($this->fetch_verify_client_result($result)) {
      $this->object->result = $result;
      $this->object->entity = (object) $result;
    }
    else {
      // TODO: give instructions to the controller to bail everything.
      // probably flag a problem and end the phase and return operation to the controller
    }
    // Check for sanity here?
    //dsm($result, 'fetch result');
    //return "node from drupal! $nid";
  }
  
  /**
   * Phase handler: node - filter.
   *
   * The filter phase performs any clean-up required according to the object's version.
   */
  function run_phase_filter() {    
    // It's up to the individual implementation to translate any field names.
    // TODO: find a way to move the alter hook up to common code,
    // maybe add a prefilter phase just for that?
    // Alternatively, add alter hooks before / after EVERY phase in the core
    // run method?
    $type = $this->object->type;
    drupal_alter('transport_' . $type . '_prefilter', $this->object);

    // Actually very little needs to be done here usually.
    // Unset the nid and fid.
    unset($this->object->entity->nid);
    unset($this->object->entity->vid);
    
    
    // DEVELOPMENT CODE!!!!! REMOVE THIS LATER TODO TODO TODO
    unset($this->object->entity->taxonomy);
  }  
}

// ======================================== Stand-in hook implementations: node
/**
 * Implementation of hook_transport_process_dependencies_node() for user module.
 *
 * Verifies a node for user dependencies from the author of the node.
 *
 * hook_transport_process_dependencies_node() should return an array of
 * dependencies of the given entity, in the form:
 * $dependencies[] = array(
 *  'type' => TYPE,
 *  'id'   => ID,
 *  'ref'  => a variable reference to the actual
 * );
 */
function user_transport_process_dependencies_node($entity_handler) {
  $remote_entity = $entity_handler->entity;

  // The author of the node is a user dependency.
  // If you don't want to bring over users, use the alter hook in the filter phase
  // to change the uid of the node to, say, 1.
  //dsm($remote_entity, 'user deps');
  
  // Users 0 and 1 are never dependencies.
  if ($remote_entity->uid == 0 || $remote_entity->uid == 1) {
    return;
  }

  $deps[] = array(
    'type'  => 'user',
    'id'    => $remote_entity->uid,
    'ref'   => &$remote_entity->uid,
  );
  
  return $deps;
}

/**
 * Implementation of hook_transport_process_dependencies_node() for taxonomy.
 */
function taxonomy_transport_process_dependencies_node($entity_handler) {
  // TODO.
}

/**
 * Implementation of hook_transport_process_dependencies_node() for upload.
 *
 * Verifies a node for file dependencies from upload.
 */
function upload_transport_process_dependencies_node($entity_handler) {
  // TODO.
}

/**
 * Implementation of hook_transport_process_dependencies_node() for content.
 *
 * Content module does not have any dependencies to add, but we use this
 * hook to add in CCK field data, as it runs before all other CCK field modules.
 *
 * TODO: consider doing this in an earlier phase, eg init?
 */
function content_transport_process_dependencies_node($entity_handler) {
  $remote_entity = $entity_handler->entity;
  
  // Stash CCK's field definitions for this node type.
  // Get CCK's data about the local node type and thence a list of its fields.
  $content_type_info = content_types($remote_entity->type);
  $fields = $content_type_info['fields'];
  
  // Restructure this array so it's grouped by module, then by field type.
  // This saves fieldtype modules the bother of having to repeat this sort of thing.
  $fields_by_module = array();
  foreach ($fields as $fieldname => $field) {
    $fields_by_module[$field['module']][$field['type']][$fieldname] = $field;
  }
  //dsm($fields_by_module, 'fields by module');
  // We may get no key for a given CCK module in $fields_by_module
  // causing an error in, eg, nodereference_transport_process_dependencies_node
  // when it tries to iterate. Up to the later hook to cast to array; nothing we can do here.
  
  $entity_handler->module_data['content']['fields'] = $fields_by_module;
}

/**
 * Implementation of hook_transport_process_dependencies_node() for nodereference.
 *
 * Verifies a node for node dependencies from nodereference.
 */
function nodereference_transport_process_dependencies_node($entity_handler) {
  $remote_entity = $entity_handler->entity;
  //dsm($remote_entity, 'noderef');
  
  // Retrieve the fieldlist from where content module's hook stored it.
  // We only have one field type anyway, so all our fields are beneath MODULE FIELDTYPE.
  $nodereferencefields = (array) $entity_handler->module_data['content']['fields']['nodereference']['nodereference'];
  // TODO: check field is set here!!!!
  foreach ($nodereferencefields as $field_name => $field) {
    //dsm($remote_entity->$field_name, 'nrf');
    if (isset($remote_entity->$field_name)) {
      foreach ($remote_entity->$field_name as $delta => $data) {
        $deps[] = array(
          'type'  => 'node',
          'id'    => $data['nid'],
          'ref'   => &$remote_entity->{$field_name}[$delta]['nid'],
        );
      }
    }
  }

  return $deps;
}


/**
 * Implementation of hook_transport_process_dependencies_node() for filefield.
 *
 * Verifies a node for file dependencies in CCK file- and imagefields.
 */
function filefield_transport_process_dependencies_node($entity_handler) {
  $remote_entity = $entity_handler->entity;
 
  // Retrieve the fieldlist from where content module's hook stored it.
  // We only have one field type anyway, so all our fields are beneath MODULE FIELDTYPE.
  $filefields = (array) $entity_handler->module_data['content']['fields']['filefield']['filefield'];
  
  // Now we have our data!
  //dsm($remote_entity, 'file deps remote ent');
  //dpr($remote_entity);
  //dsm($filefields);
  
  foreach ($filefields as $field_name => $field) {
    // Check for both existence of the local field on the remote node, and
    // whether there is any data in it; CCK 6 at least has a (mostly) empty 
    // 0 key even if there is no data.
    if (isset($remote_entity->$field_name) && is_array($remote_entity->{$field_name}[0])) {
      foreach ($remote_entity->$field_name as $delta => $data) {
        $deps[] = array(
          'type' => 'file',
          'id'   => $data['fid'],
          'ref'  => &$remote_entity->{$field_name}[$delta]['fid'],
          // Files are a somewhat special case, as to save a file we need the
          // field definition, which by the time we'll be in the remote file
          // entity handler we won't have.
          // So we stash it here.
          // This may well break if the same fid is put into dependencies
          // but with filefield 5 that's REALLY unlikely.
          'data' => array(
            'file'      => $data, 
            'filefield' => $field,
          ),
        );
        
      }      
    }
  }
  
  //dsm($remote_entity, 'done file deps');
  return $deps;  
}

/**
 * Implementation of hook_transport_process_mappings_node() for node.
 *
 * Add a node's format property to the list of properties to map.
 */
function node_transport_process_mappings_node($entity_handler) {
  //dsm($entity_handler);
  $remote_entity = $entity_handler->entity;
  $mapping_list[] = array(
    'group' => 'input_format',
    'id'    => $remote_entity->format,
    'ref'   => &$remote_entity->format,
  );
  return $mapping_list;
}

/**
 * Implementation of hook_transport_process_mappings_node() for (CCK) text.
 *
 * Add the format properties for any CCK text fields that have them to 
 * the list of properties to map.
 */
function text_transport_process_mappings_node($entity_handler) {
  $remote_entity = $entity_handler->entity;
  
  // Remember that field data stash is grouped by CCK module first, THEN
  // field type to cover the (non-core AFAIK) cases where one module
  // has several types. Hence the slightly WTF double 'text'.
  // The other WTF is that it is essential to cast to an array here.
  $text_fields = (array) $entity_handler->module_data['content']['fields']['text']['text'];
  
  // From text.module:
  // $options = array(0 => t('Plain text'), 1 => t('Filtered text (user selects input format)'));
  foreach ($text_fields as $field_name => $field) {
    // Only act on text fields set to have an input format, ie not plain text.
    if ($field['text_processing'] == 1) {
      // I wish PHP were smart enough to treat non-existent things as an empty array :/
      if (isset($remote_entity->$field_name) && is_array($remote_entity->$field_name)) {
        foreach ($remote_entity->$field_name as $delta => $data) {
          // We have to add every item to the mapping properties, as each one
          // has its own format.
          $mapping_list[] = array(
            'group' => 'input_format',
            'id'    => $remote_entity->{$field_name}[$delta]['format'],
            'ref'   => &$remote_entity->{$field_name}[$delta]['format'],
          );
        }
      }
    }
  }

  return $mapping_list;  
}

// ======================================== Extender classes: File

/**
 * Entity class for files.
 *
 * The entity class should contain methods for phases that do not have interfaces;
 * ie those that are identical no matter the source.
 *
 * WARNING: some file API functions expect files to be arrays of data, but 
 * within the Remote entity handler we're storing an object in $handler->entity,
 * hence this will require casting to array in some cases.
 *
 * Files are only retrieved once; if a file exists locally then it is skipped.
 * (This is based on the remote fid, rather than the filename which could
 * merely be a coincidence.)
 * This is because Drupal core and filefield never change
 * a file or its {files} table record once it is saved, AFAIK.
 * (Please file an issue if there are special cases!)
 */
class TransportRemoteEntity_file extends TransportRemoteEntity {
  /**
   * Phase handler: file - init.
   */
  function run_phase_init() {
    // It is extremely important to call the parent method here.
    parent::run_phase_init();
    
    // Skip if this file has already been retrieved.
    if ($this->local_exists) {
      $this->advance_phase = FALSE;
      $this->status = 'skip';
      //dsm($this, 'skip file');
    }
  }
  
  /**
   * Phase handler: file - dependencies.
   *
   * Files don't have dependencies, at least not on D5.
   * The uid in theory should already have been brought over -- we hope!
   * Arg that means we may need to version this too????
   */  
  function run_phase_dependencies() {
  }
  
  /**
   * Phase handler: file - translate.
   */  
  function run_phase_translate() {
    // TODO: translate file uid property.
  }
  
  /**
   * Phase handler: file - save.
   *
   * Oh argh. This may have to be versioned for files :(
   * Which means phases are versioned inconsistently per type... fun!
   */  
  function run_phase_save() {
    //dsm($this, 'phase save file');
        
    // Some terminology lest I lose my mind:
    //  - $file: the array that things like drupal_write_record and cck 
    //      filefield saving will expect. Contains data like fid, title,
    //      alt, nid, filemime and all that jazz.
    //  - $file_data: the actual raw file.
    //  - $field: the filefield array from CCK, in the event we're a CCK
    //      file. 
    //      TODO: when we come to handle upload files too, this will BREAK! ;)
    $file  = $this->entity;
    $file_data = $this->entity->file;
    $field = $this->data['data']['filefield'];
    
    // Create the filepath to save to.
    // TODO: brittle! account for upload module too!
    $file->filepath = filefield_widget_file_path($field) . '/' . $file->filename;
    
    // ALSO need data that was stored on the CCK filefield when it came over with the node!
    // Do we? what for???
    $filefield_file  = $this->data['data']['file'];
    
    // In all cases, remove the fid that came over with the old node.
    unset($file->fid);

    // Save the incoming file.
    // Gah this is so bloody convoluted. Roll on D7.
    // ALSO! This fails if the local site has not yet saved a file to the
    // current filefield and it has a folder to itself -- because the folder
    // doesn't exist yet!!!!!!
    // If a file already existed of the same name it's merely a coincidence, so
    // we use FILE_EXISTS_RENAME to avoid clobbering it.
    $filepath = file_save_data($file_data, $file->filepath, FILE_EXISTS_RENAME);
    $file->filepath = $filepath;
    
    // Write record in files table.
    drupal_write_record('files', $file);      
    $this->id_local = $file->fid;
    
    //dsm($file, 'file saved');
    
    // Put the file back in the data... though DUH we should be using $this->entity,
    // it's what it's there for!
    $this->data['data']['file'] = $file;    
  }  
}

/**
 * Extender Class: fetch phase.
 *
 * Type:   File.
 * Source: Clients Drupal 5
 */
class TransportRemoteEntityExtender_fetch_file_d5 
  extends TransportRemoteEntityExtender_fetch_client 
  implements TransportRemoteEntityInterface_fetch {

  /**
   * Phase handler: file - fetch - D5.
   *
   * Fetch a file from a D5 site. 
   * Because there's no file service on D5 Services, so we'd need to backport
   * the module at http://drupal.org/node/290053 and bundle it in to stuff
   * we're providing for D5.
   * In the meantime, the direct approach is fine if all the files are public, 
   * but this needs fixing if they are not: TODO.
   */
  function run_phase_fetch() {
    //dsm($this->object->source);
    
    // Figure out the base URL of the remote site based on
    $remote_url = substr($this->object->source->endpoint, 0, -15);
    
    // Retrieve the filefield data that we sneakily stashed when we came here for dependencies...
    // of course this will completely fail when we're here for upload module files
    // TODO: write a better way. See the TODO at the top of this method.
    $file = $this->object->data['data']['file'];
    
    // Note we have no timestamps coming in from D5.
    $file_url = $remote_url . '/' . $file['filepath'];

    // The URL may have spaces in and PHP is too stupid to figure this out.
    $file_url = str_replace(' ', '%20', $file_url);

    //dsm($file_url);
    $file_data = file_get_contents($file_url);
    if ($file_data === FALSE) {
      // an error happened
    }
    
    // D6 services puts the actual file data here so we will too.
    $file['file'] = $file_data;
    
    // Save the ersatz file array as the remote entity.
    $this->object->entity = (object) $file;
    
    // shortly TODO obsolete
    $this->object->file_data = $file_data;
    //file_put_contents('pic', $file_data);
  }

  /**
   * Phase handler: file - filter - D5
   */
  function run_phase_filter() {
    // Change the  of properties from D5 to D6 style should happen!
    // And obviously, run the alter hook...
    
    //dsm($this->object, 'filter d5');
    $entity_handler = $this->object;
    
    // Filefield on D5 stores title and alt attributes in the base file array;
    // these must be moved to a subarray 'data' for D6.
    $file = $entity_handler->data['data']['file']; // Just a local copy for sanity.
    $entity_handler->data['data']['file']['data'] = array(
      'title' => $file['title'],
      'alt'   => $file['alt'],
    );
    
    // Other things that D6 expects to have...
    $entity_handler->data['data']['file']['list'] = 1; // well this one is for filefield...
    $entity_handler->data['data']['file']['status'] = 1;
    // TODO!!!!
    $entity_handler->data['data']['file']['uid'] = 1;
    // Note also that D5 files don't have a timestamp, but that should be ok.
    
    
    $entity_handler->entity->status = 1;
    // TODO!!!! Set the actual uid from the filefield data!
    $entity_handler->entity->uid = 1;
    
  }
}

/**
 * Extender Class: fetch phase.
 *
 * Type:   File.
 * Source: Clients Drupal 6 Services 2
 */
class TransportRemoteEntityExtender_fetch_file_d6_2 
  extends TransportRemoteEntityExtender_fetch_client 
  implements TransportRemoteEntityInterface_fetch {

  /**
   * Phase handler: file - fetch - D6.
   *
   * Fetch a file from a D6 site. 
   */
  function run_phase_fetch() {
    
    // Something in the Services and XMLRPC system is REALLY fussy
    // about strings vs ints, so we have to cast it.
    $fid = (int) $this->object->get_id();
    $result = $this->object->source->callMethod('file.get', $fid);
    if ($this->fetch_verify_client_result($result)) {
      $this->object->result = $result;
      $this->object->entity = (object) $result;
      
    }
  }

  /**
   * Phase handler: file - filter - D6.
   */
  function run_phase_filter() {
    // Drupal 6 Services encodes file data...
    $this->object->entity->file = base64_decode($this->object->entity->file);
  }

}

// ======================================== Extender classes: user

/**
 * Entity class for users.
 *
 * The entity class should contain methods for phases that do not have interfaces;
 * ie those that are identical no matter the source.
 */
class TransportRemoteEntity_user extends TransportRemoteEntity {
  /**
   * Phase handler: user - init.
   */
  function run_phase_init() {
    // It is extremely important to call the parent method here.
    parent::run_phase_init();

    // Skip handling of users 0 and 1 from the remote site.
    if ($this->data['id'] == 0 || $this->data['id'] == 1) {
      $this->advance_phase = FALSE;
      $this->status = 'skip';
      //dsm($this, 'skip user');
    }
  }
  
  /**
   * Phase handler: user - dependencies.
   * 
   * Users don't have dependencies -- at least in core D6.
   */  
  function run_phase_dependencies() {
    // TODO: Invoke the hook anyway as other modules may create dependencies.
  }
  
  /**
   * Phase handler: user - translate.
   */  
  function run_phase_translate() {
    // Nothing to do: no dependencies. 
    // TODO: Invoke the hook anyway as other modules may create dependencies.
  }
  
  /*
  !! TODO: need an extra phase: 'preexisting' --
  for users and files, (and terms?) check for an existing, identical entity.
  ie for user, check email
  files, filename + node
  etc
  */
  
  /**
   * Phase handler: user - save.
   */  
  function run_phase_save() {
    // user_save() is WEIRD.
    // TODO: clean up this code.
    
    $remote_user = $this->entity;
    //dsm($remote_user, 'rem user');
    
    if ($this->local_exists) {
      $local_user = user_load($this->id_local);

      unset($remote_user->uid);
      
      $user_array = (array) $remote_user;
      $new_user = user_save($local_user, $user_array);
    }
    else {
      unset($remote_user->uid);
      
      $user_array = (array) $remote_user;
      $new_user = user_save('', $user_array);
      
      $this->id_local = $new_user->uid;
      //dsm($new_user);

      // Password would have been doubly md5'ed, so set explicitly
      db_query("UPDATE {users} SET pass = '%s' WHERE uid = %d", $user_array['pass'], $new_user->uid);
    }
  }
}

/**
 * Extender Class: fetch phase.
 *
 * Type:   User.
 * Source: Clients Drupal 5
 */
class TransportRemoteEntityExtender_fetch_user_d5 
  extends TransportRemoteEntityExtender_fetch_client 
  implements TransportRemoteEntityInterface_fetch {

  /**
   * Phase handler: user - fetch - D5.
   *
   * Fetch a user from a D5 site. 
   */
  function run_phase_fetch() {
    // Something in the Services and XMLRPC system is REALLY fussy
    // about strings vs ints, so we have to cast it.
    $uid = (int) $this->object->get_id();
    $result = $this->object->source->callMethod('user.load', $uid);
    //dpm($result, 'user 5 fetch');
    if ($this->fetch_verify_client_result($result)) {
      $this->object->result = $result;
      $this->object->entity = (object) $result;
    }
    else {
      // probably flag a problem and end the phase and return operation to the controller
    }
  }
  
  /**
   * Phase handler: user - filter - D5.
   */
  function run_phase_filter() {
    // Run the alter hook first so modules get the raw data.
    drupal_alter('transport_preprocess_' . $this->object->type, $this->object->entity);
    
    // We don't handle roles yet, so clear them as this could be a
    // security issue.
    $this->object->entity->roles = array();
    
    // TODO: Clear/check 'theme' property.
    
    //dsm($this->object->entity);
  }  
}
